@@ -14,7 +14,7 @@ tvROC.jm <- function (object, newdata, Tstart, Thoriz = NULL, Dt = NULL, ...) {
     if (is.null(Thoriz))
         Thoriz <- Tstart + Dt
     type_censoring <- object$model_info$type_censoring
-    if (type_censoring != "right")
+    if (object$model_info$CR_MS)
         stop("'tvROC()' currently only works for right censored data.")
     Tstart <- Tstart + 1e-06
     Thoriz <- Thoriz + 1e-06
@@ -26,22 +26,24 @@ tvROC.jm <- function (object, newdata, Tstart, Thoriz = NULL, Dt = NULL, ...) {
         stop("cannot find the '", id_var, "' variable in newdata.", sep = "")
     if (is.null(newdata[[time_var]]))
         stop("cannot find the '", time_var, "' variable in newdata.", sep = "")
-    if (is.null(newdata[[Time_var]]))
-        stop("cannot find the '", Time_var, "' variable in newdata.", sep = "")
+    if (any(sapply(Time_var, function (nmn) is.null(newdata[[nmn]]))))
+        stop("cannot find the '", paste(Time_var, collapse = ", "),
+             "' variable(s) in newdata.", sep = "")
     if (is.null(newdata[[event_var]]))
         stop("cannot find the '", event_var, "' variable in newdata.", sep = "")
-    newdata <- newdata[newdata[[Time_var]] > Tstart, ]
+    tt <- if (type_censoring == "right") newdata[[Time_var]] else newdata[[Time_var[2L]]]
+    newdata[[id_var]] <- newdata[[id_var]][, drop = TRUE]
+    id <- newdata[[id_var]]
+    id <- match(id, unique(id))
+    tt <- ave(tt, id, FUN = function (t) rep(tail(t, 1L) > Tstart, length(t)))
+    newdata <- newdata[as.logical(tt), ]
     newdata <- newdata[newdata[[time_var]] <= Tstart, ]
     if (!nrow(newdata))
         stop("there are no data on subjects who had an observed event time after Tstart",
              "and longitudinal measurements before Tstart.")
-    newdata[[id_var]] <- newdata[[id_var]][, drop = TRUE]
     test1 <- newdata[[Time_var]] < Thoriz & newdata[[event_var]] == 1
     if (!any(test1))
         stop("it seems that there are no events in the interval [Tstart, Thoriz).")
-    #test2 <- newdata[[Time_var]] > Thoriz & newdata[[event_var]] == 1
-    #if (!any(test2))
-    #    stop("it seems that there are no events after Thoriz.")
     newdata2 <- newdata
     newdata2[[Time_var]] <- Tstart
     newdata2[[event_var]] <- 0
@@ -159,7 +161,7 @@ tvAUC.jm <- function (object, newdata, Tstart, Thoriz = NULL, Dt = NULL, ...) {
     if (is.null(Thoriz))
         Thoriz <- Tstart + Dt
     type_censoring <- object$model_info$type_censoring
-    if (type_censoring != "right")
+    if (object$model_info$CR_MS)
         stop("'tvROC()' currently only works for right censored data.")
     Tstart <- Tstart + 1e-06
     Thoriz <- Thoriz + 1e-06
@@ -171,8 +173,9 @@ tvAUC.jm <- function (object, newdata, Tstart, Thoriz = NULL, Dt = NULL, ...) {
         stop("cannot find the '", id_var, "' variable in newdata.", sep = "")
     if (is.null(newdata[[time_var]]))
         stop("cannot find the '", time_var, "' variable in newdata.", sep = "")
-    if (is.null(newdata[[Time_var]]))
-        stop("cannot find the '", Time_var, "' variable in newdata.", sep = "")
+    if (any(sapply(Time_var, function (nmn) is.null(newdata[[nmn]]))))
+        stop("cannot find the '", paste(Time_var, collapse = ", "),
+             "' variable(s) in newdata.", sep = "")
     if (is.null(newdata[[event_var]]))
         stop("cannot find the '", event_var, "' variable in newdata.", sep = "")
     newdata <- newdata[order(newdata[[Time_var]]), ]
@@ -182,9 +185,6 @@ tvAUC.jm <- function (object, newdata, Tstart, Thoriz = NULL, Dt = NULL, ...) {
     test1 <- newdata[[Time_var]] < Thoriz & newdata[[event_var]] == 1
     if (!any(test1))
         stop("it seems that there are no events in the interval [Tstart, Thoriz).")
-    #test2 <- newdata[[Time_var]] > Thoriz & newdata[[event_var]] == 1
-    #if (!any(test2))
-    #    stop("it seems that there are no events after Thoriz.")
     newdata2 <- newdata
     newdata2[[Time_var]] <- Tstart
     newdata2[[event_var]] <- 0
@@ -331,7 +331,7 @@ calibration_plot <- function (object, newdata, Tstart, Thoriz = NULL,
     if (is.null(Thoriz))
         Thoriz <- Tstart + Dt
     type_censoring <- object$model_info$type_censoring
-    if (type_censoring != "right")
+    if (object$model_info$CR_MS)
         stop("'tvROC()' currently only works for right censored data.")
     Tstart <- Tstart + 1e-06
     Thoriz <- Thoriz + 1e-06
@@ -343,8 +343,9 @@ calibration_plot <- function (object, newdata, Tstart, Thoriz = NULL,
         stop("cannot find the '", id_var, "' variable in newdata.", sep = "")
     if (is.null(newdata[[time_var]]))
         stop("cannot find the '", time_var, "' variable in newdata.", sep = "")
-    if (is.null(newdata[[Time_var]]))
-        stop("cannot find the '", Time_var, "' variable in newdata.", sep = "")
+    if (any(sapply(Time_var, function (nmn) is.null(newdata[[nmn]]))))
+        stop("cannot find the '", paste(Time_var, collapse = ", "),
+             "' variable(s) in newdata.", sep = "")
     if (is.null(newdata[[event_var]]))
         stop("cannot find the '", event_var, "' variable in newdata.", sep = "")
     newdata <- newdata[newdata[[Time_var]] > Tstart, ]
@@ -356,9 +357,6 @@ calibration_plot <- function (object, newdata, Tstart, Thoriz = NULL,
     test1 <- newdata[[Time_var]] < Thoriz & newdata[[event_var]] == 1
     if (!any(test1))
         stop("it seems that there are no events in the interval [Tstart, Thoriz).")
-    #test2 <- newdata[[Time_var]] > Thoriz & newdata[[event_var]] == 1
-    #if (!any(test2))
-    #    stop("it seems that there are no events after Thoriz.")
     newdata2 <- newdata
     newdata2[[Time_var]] <- Tstart
     newdata2[[event_var]] <- 0
@@ -428,7 +426,7 @@ tvBrier <- function (object, newdata, Tstart, Thoriz = NULL, Dt = NULL, ...) {
     if (is.null(Thoriz))
         Thoriz <- Tstart + Dt
     type_censoring <- object$model_info$type_censoring
-    if (type_censoring != "right")
+    if (object$model_info$CR_MS)
         stop("'tvROC()' currently only works for right censored data.")
     Tstart <- Tstart + 1e-06
     Thoriz <- Thoriz + 1e-06
@@ -440,8 +438,9 @@ tvBrier <- function (object, newdata, Tstart, Thoriz = NULL, Dt = NULL, ...) {
         stop("cannot find the '", id_var, "' variable in newdata.", sep = "")
     if (is.null(newdata[[time_var]]))
         stop("cannot find the '", time_var, "' variable in newdata.", sep = "")
-    if (is.null(newdata[[Time_var]]))
-        stop("cannot find the '", Time_var, "' variable in newdata.", sep = "")
+    if (any(sapply(Time_var, function (nmn) is.null(newdata[[nmn]]))))
+        stop("cannot find the '", paste(Time_var, collapse = ", "),
+             "' variable(s) in newdata.", sep = "")
     if (is.null(newdata[[event_var]]))
         stop("cannot find the '", event_var, "' variable in newdata.", sep = "")
     newdata <- newdata[newdata[[Time_var]] > Tstart, ]
@@ -453,9 +452,6 @@ tvBrier <- function (object, newdata, Tstart, Thoriz = NULL, Dt = NULL, ...) {
     test1 <- newdata[[Time_var]] < Thoriz & newdata[[event_var]] == 1
     if (!any(test1))
         stop("it seems that there are no events in the interval [Tstart, Thoriz).")
-    #test2 <- newdata[[Time_var]] > Thoriz & newdata[[event_var]] == 1
-    #if (!any(test2))
-    #    stop("it seems that there are no events after Thoriz.")
     newdata2 <- newdata
     newdata2[[Time_var]] <- Tstart
     newdata2[[event_var]] <- 0
