@@ -144,6 +144,7 @@ jm_fit <- function (model_data, model_info, initial_values, priors, control) {
         colnames(out[[i]][["mcmc"]][["sigmas"]]) <-
             paste0("sigmas_",
                    seq_along(out[[i]][["mcmc"]][["sigmas"]][1, ]))
+        colnames(out[[i]][["mcmc"]][["sigmaF"]]) <- "sigma_frailty"
     }
     # drop sigmas that are not needed
     has_sigmas <- initial_values$log_sigmas > -20.0
@@ -167,9 +168,9 @@ jm_fit <- function (model_data, model_info, initial_values, priors, control) {
     }
     parms <- c("bs_gammas", "tau_bs_gammas", "gammas", "alphas", "W_std_gammas",
                "Wlong_std_alphas", "D", paste0("betas", seq_along(model_data$X)),
-               "sigmas")
+               "sigmas", "alphaF", "frailty", "sigmaF")
     parms2 <- c("bs_gammas", "gammas", "alphas", "L", "sds", "betas", "b",
-                "sigmas")
+                "sigmas", "alphaF", "frailty", "sigmaF")
     if (control$save_random_effects) parms <- c(parms, "b")
     if (!length(attr(model_info$terms$terms_Surv_noResp, "term.labels")))
         parms <- parms[parms != "gammas"]
@@ -287,10 +288,13 @@ jm_fit <- function (model_data, model_info, initial_values, priors, control) {
     res_thetas[["betas"]] <-
         lapply(mcmc_out$mcmc[grep("^betas", names(mcmc_out$mcmc))], function (m)
             do.call("rbind", m))
+    res_thetas$frailty <- do.call("rbind", mcmc_out$mcmc$frailty)
+    res_thetas$alphaF  <- do.call("rbind", mcmc_out$mcmc$alphaF)
+    res_thetas$sigmaF  <- do.call("rbind", mcmc_out$mcmc$sigmaF)
     mcmc_out$mlogLik <- mlogLik_jm(res_thetas, statistics$Mean[["b"]],
                           statistics$post_vars, model_data, model_info, control)
     ind <- names(thetas) %in% c("sigmas", "bs_gammas", "gammas", "alphas",
-                                "tau_bs_gammas")
+                                "tau_bs_gammas", "alphaF", "frailty", "sigmaF")
     thetas[ind] <- lapply(thetas[ind], rbind)
     thetas$betas <- lapply(thetas$betas, rbind)
     dim(thetas[["D"]]) <- c(dim(thetas[["D"]]), 1L)
